###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import ast
import subprocess
import logging
from collections import defaultdict

import unitem.seq_io as seq_io

from unitem.defaults import *


def run_cmd(cmd, program=None):
    """Run external command."""

    logger = logging.getLogger('timestamp')
    logger.info(f"Executing: {cmd}")

    try:
        proc = subprocess.Popen(cmd.split(),
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                encoding='utf-8')

        progress_strs = ['Finished processing',
                         'Finished parsing']
        while True:
            out = proc.stdout.readline()
            if not out and proc.poll() is not None:
                break

            if out.rstrip():
                # HACK: if output is logging information for a dependency, remove
                # the logging information so it isn't shown twice
                if ' INFO: ' in out:
                    out = out.split(' INFO: ')[1]

                if any([p in out for p in progress_strs]):
                    # HACK: CheckM progress bars so write to stdout without newline
                    print(f'{out.rstrip()}\r', end='')
                elif program:
                    logger.info(f'[{program}] {out.rstrip()}')
                else:
                    logger.info(out.rstrip())

        if proc.returncode != 0:
            logger.error(f'Return code: {proc.returncode}')
            sys.exit(1)
    except Exception as e:
        print(e)
        logger.error('Failed to execute command.')
        sys.exit(1)


def calculateN50L50M50(seqs):
    """Calculate N50 and M50 statistics."""

    seq_lens = [len(seq) for seq in seqs]

    thresholdN50 = sum(seq_lens) / 2

    seq_lens.sort(reverse=True)

    test_sum = 0
    L50 = 0
    N50 = 0
    for seq_len in seq_lens:
        L50 += 1
        test_sum += seq_len
        if test_sum >= thresholdN50:
            N50 = seq_len
            break

    M50 = len(seqs) - L50
    if test_sum != thresholdN50:
        M50 += 1

    return N50, L50, M50


def parse_checkm_bin_stats(checkm_dir):
    """Read bin statistics from file."""
    bin_stats_file = os.path.join(
        checkm_dir, 'storage', BIN_STATS_EXT_OUT)

    bin_stats = {}
    with open(bin_stats_file, 'r') as f:
        for line in f:
            line_split = line.split('\t')
            bin_stats[line_split[0]] = ast.literal_eval(line_split[1])

    return bin_stats


def parse_bin_stats(profile_dir):
    """Parse genomic and assembly statistics for bins."""

    binning_methods_dir = os.path.join(profile_dir, BINNING_METHOD_DIR)
    bin_stats = {}
    for bm in os.listdir(binning_methods_dir):
        checkm_dir = os.path.join(binning_methods_dir, bm, CHECKM_BAC_DIR)

        bin_stats[bm] = parse_checkm_bin_stats(checkm_dir)

    return bin_stats


def read_bins(bin_dirs):
    """Read sequences in bins."""

    bins = defaultdict(lambda: defaultdict(set))
    contigs = {}
    contigs_in_bins = defaultdict(lambda: {})
    for method_id, (bin_dir, bin_ext) in bin_dirs.items():
        for bf in os.listdir(bin_dir):
            if not bf.endswith(bin_ext):
                continue

            bin_id = bf[0:bf.rfind(bin_ext)]
            if bin_id[-1] == '.':
                bin_id = bin_id[0:-1]
            bf_path = os.path.join(bin_dir, bf)

            for seq_id, seq in seq_io.read_seq(bf_path):
                bins[method_id][bin_id].add(seq_id)
                contigs[seq_id] = seq
                contigs_in_bins[seq_id][method_id] = bin_id

            if len(bins[method_id][bin_id]) == 0:
                logger = logging.getLogger('timestamp')
                logger.warning('Bin {bf} from {method_id} is empty.')

    return bins, contigs, contigs_in_bins


def count_nt(seq):
    """Count occurrences of each nucleotide in a sequence.

    Only the bases A, C, G, and T(U) are counted. Ambiguous
    bases are ignored.

    Parameters
    ----------
    seq : str
        Nucleotide sequence.

    Returns
    -------
    list
        Number of A, C, G, and T(U) in sequence.
    """

    s = seq.upper()
    a = s.count('A')
    c = s.count('C')
    g = s.count('G')
    t = s.count('T') + s.count('U')

    return a, c, g, t


def bin_gc(bin):
    """Calculate GC-content of bin."""

    a_count = 0
    c_count = 0
    g_count = 0
    t_count = 0
    for seq in bin.values():
        a, c, g, t = count_nt(seq)

        a_count += a
        c_count += c
        g_count += g
        t_count += t

    total = (a_count + c_count + g_count + t_count)

    return (g_count + c_count) / total



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
from typing import Tuple, Optional, Iterator

from unitem.utils import make_sure_path_exists
from unitem.defaults import TIGRFAM_TOP_HIT_SUFFIX, PFAM_TOP_HIT_SUFFIX


class Hit(object):
    """More significant hits have a higher bit-score and lower e-value.

    Running sorted() on a list of Hit objects will yield a list in order of
    least significant to most significant hits.
    """

    def __init__(self, gene_id: str, hmm_id: str, e_val: float, bit_score: float):
        """Store information about this hit."""

        self.gene_id = gene_id
        self.hmm_id = hmm_id
        self.e_val = e_val
        self.bit_score = bit_score

    def __repr__(self) -> str:

        return f'{self.gene_id} {self.hmm_id} ({self.e_val}/{self.bit_score})'

    def __eq__(self, other) -> bool:
        """Returns True if self is equal to other."""

        return isinstance(other, Hit) and other.gene_id == self.gene_id \
            and other.hmm_id == self.hmm_id and other.e_val == self.e_val \
            and other.bit_score == self.bit_score

    def __lt__(self, other) -> bool:
        """Is self less significant than other?"""
        if self.bit_score < other.bit_score:
            return True
        elif self.bit_score == other.bit_score:
            if self.e_val > other.e_val:
                return True
            elif self.e_val == other.e_val:
                if self.hmm_id > other.hmm_id:
                    return True
                elif self.hmm_id == other.hmm_id:
                    return self.gene_id > other.gene_id
        return False

    def __gt__(self, other) -> bool:
        """Is self more significant than other?"""

        raise NotImplemented

    def __hash__(self) -> int:

        return hash(f'{self.gene_id}_{self.hmm_id}_{self.e_val}_{self.bit_score}')

    def hmm_str(self) -> str:
        """Report the e-value and bit-score for the hmm."""

        return f'{self.hmm_id},{self.e_val},{self.bit_score}'


class TopHitFile(object):
    """Stores information about the top marker hits for Pfam/TIGRFAM markers."""

    def __init__(self, path: str):
        """Setup the path to the file and initialise storage dictionary."""

        self.path = path
        self.hits = dict()

    def add_hit(self, gene_id: str, hmm_id: str, e_val: float, bit_score: float):
        """Store the most significant HMM hit for each gene."""

        if gene_id not in self.hits:
            self.hits[gene_id] = dict()

        # Check if this hit already exists.
        new_hit = Hit(gene_id, hmm_id, e_val, bit_score)
        if hmm_id in self.hits[gene_id]:
            # Store the new hit if it's more significant.
            if self.hits[gene_id][hmm_id] < new_hit:
                self.hits[gene_id][hmm_id] = new_hit
        else:
            self.hits[gene_id][hmm_id] = new_hit

    def contains_gene_id(self, gene_id: str) -> bool:
        """Returns True if the gene_id is found in the top hit file."""

        return gene_id in self.hits

    def contains_gene_hmm(self, gene_id: str, hmm_id: str) -> bool:
        """Returns True if there is a hmm hit for the gene_id."""

        return gene_id in self.hits and hmm_id in self.hits[gene_id]

    def get_top_hit(self, gene_id: str) -> Optional[Hit]:
        """Returns the most significant hit for a given gene id or None."""

        if gene_id not in self.hits or len(self.hits[gene_id]) == 0:
            return None
        return sorted(self.hits[gene_id].values(), reverse=True)[0]

    def get_hmm_hit(self, gene_id: str, hmm_id: str) -> Hit:
        """Returns the hit indexed by the gene and hmm id."""

        return self.hits[gene_id][hmm_id]

    def write(self):
        """Writes the file to disk."""

        make_sure_path_exists(os.path.dirname(self.path))
        header = ['Gene Id', 'Top hits (Family id,e-value,bitscore)']
        with open(self.path, 'w') as fh:
            fh.write('\t'.join(header) + '\n')
            for gene_id, hits in sorted(self.hits.items()):
                out_hits = list()
                for cur_hit in sorted(hits.values(), reverse=True):
                    out_hits.append(cur_hit.hmm_str())
                concat_hits = ';'.join(out_hits)
                fh.write(f'{gene_id}\t{concat_hits}\n')

    def read(self):
        """Read the contents of an existing tophit file."""

        with open(self.path) as fh:
            fh.readline()
            for line in fh:
                gene_id, hits = line.strip().split('\t')
                for hit in hits.split(';'):
                    hmm_id, e_val, bit_score = hit.split(',')
                    e_val, bit_score = float(e_val), float(bit_score)
                    self.add_hit(gene_id, hmm_id, e_val, bit_score)

    def iter_hits(self) -> Iterator[Tuple[str, Hit]]:
        """Iterate over all genes and their respective hits."""

        for gene_id, hmm_dict in self.hits.items():
            for cur_hit in hmm_dict.values():
                yield gene_id, cur_hit


class TopHitPfamFile(TopHitFile):
    """The top hit file given the Pfam output."""

    def __init__(self, out_dir: str, gid: str):
        path = os.path.join(out_dir, f'{gid}{PFAM_TOP_HIT_SUFFIX}')
        super().__init__(path)


class TopHitTigrFile(TopHitFile):
    """The top hit file given the Tigrfam output."""

    def __init__(self, out_dir: str, gid: str):
        path = os.path.join(out_dir, f'{gid}{TIGRFAM_TOP_HIT_SUFFIX}')
        super().__init__(path)

    def add_hit(self, gene_id: str, hmm_id: str, e_val: float, bit_score: float):
        """Only store the most significant HMM hit per gene."""

        if self.contains_gene_id(gene_id):
            # The new hit was more significant, remove any other hits.
            if self.get_top_hit(gene_id) < Hit(gene_id, hmm_id, e_val, bit_score):
                self.hits[gene_id] = dict()
                super().add_hit(gene_id, hmm_id, e_val, bit_score)
        else:
            super().add_hit(gene_id, hmm_id, e_val, bit_score)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import gzip
from collections import defaultdict

from unitem.seq_io import read_fasta


class BinTools():
    """Functions for exploring and modifying bins."""

    def __init__(self, threads=1):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

    def bin_id_from_filename(self, filename):
        """Extract bin id from bin filename."""

        bin_id = os.path.basename(filename)
        bin_id = os.path.splitext(bin_id)[0]

        return bin_id

    def _read_seq_ids(self, bin_files):
        """Read sequence IDs of all bin files."""

        bins = {}
        for bin_file in bin_files:
            bin_id = self.bin_id_from_filename(bin_file)
            bins[bin_id] = set(read_fasta(bin_file).keys())

        return bins

    def bin_files(self, bin_dir, bin_extension):
        """Get bins in directory based on extension."""

        bin_files = []
        if bin_dir is not None:
            all_files = os.listdir(bin_dir)
            for f in all_files:
                if f.endswith(bin_extension):
                    bin_file = os.path.join(bin_dir, f)
                    if os.stat(bin_file).st_size == 0:
                        self.logger.warning(
                            "Skipping bin %s as it has a size of 0 bytes." % f)
                    else:
                        bin_files.append(bin_file)

        if not bin_files:
            self.logger.error(
                "No bins found. Check the extension (-x) used to identify bins.")
            sys.exit(1)

        return sorted(bin_files)

    def unique(self, bin_files):
        """Check if sequences are assigned to multiple bins."""

        # read sequence IDs from all bins,
        # while checking for duplicate sequences within a bin
        bin_seqs = {}
        for f in bin_files:
            bin_id = self.bin_id_from_filename(f)

            if f.endswith('.gz'):
                open_file = gzip.open
            else:
                open_file = open

            seq_ids = set()
            for line in open_file(f):
                if line[0] == '>':
                    seq_id = line[1:].split(None, 1)[0]

                    if seq_id in seq_ids:
                        print('Sequence %s found multiple times in bin %s.' %
                              (seq_id, bin_id))
                    seq_ids.add(seq_id)

            bin_seqs[bin_id] = seq_ids

        # check for sequences assigned to multiple bins
        print()
        bDuplicates = False
        bin_ids = bin_seqs.keys()
        for i in range(0, len(bin_ids)):
            for j in range(i + 1, len(bin_ids)):
                seq_inter = set(bin_seqs[bin_ids[i]]).intersection(
                    set(bin_seqs[bin_ids[j]]))

                if len(seq_inter) > 0:
                    bDuplicates = True
                    print('Sequences shared between %s and %s: ' %
                          (bin_ids[i], bin_ids[j]))
                    for seq_id in seq_inter:
                        print('  ' + seq_id)
                    print()

        if not bDuplicates:
            print('No sequences assigned to multiple bins.')

    def _binning_stats(self, bins, seq_lens):
        """Calculate binning stats for a set of bins."""

        total_uniq_binned_seqs = 0
        tota_uniq_binned_bases = 0

        bin_stats = {}
        processed_seqs = set()
        repeats = set()
        for bin_id, seqs in bins.items():
            num_binned_bases = 0
            for seq_id in seqs:
                num_binned_bases += seq_lens[seq_id]
                if seq_id not in processed_seqs:
                    processed_seqs.add(seq_id)
                    tota_uniq_binned_bases += seq_lens[seq_id]
                    total_uniq_binned_seqs += 1
                else:
                    repeats.add(seq_id)

            bin_stats[bin_id] = [len(seqs), num_binned_bases]

        return bin_stats, total_uniq_binned_seqs, tota_uniq_binned_bases, len(repeats)

    def compare(self, bin_files1, bin_files2, assembly_file, output_file):
        """Compare bins from two different binning methods."""

        # determine total number of sequences
        self.logger.info('Reading bins.')
        seqs = read_fasta(assembly_file)

        seq_lens = {}
        total_bases = 0
        num_seq1K = 0
        total_bases1K = 0
        num_seq5K = 0
        total_bases5K = 0
        for seq_id, seq in seqs.items():
            seq_len = len(seq)
            seq_lens[seq_id] = seq_len
            total_bases += seq_len
            if seq_len >= 1000:
                num_seq1K += 1
                total_bases1K += seq_len
            if seq_len >= 5000:
                num_seq5K += 1
                total_bases5K += seq_len

        # determine sequences in each bin
        bins1 = self._read_seq_ids(bin_files1)
        bins2 = self._read_seq_ids(bin_files2)

        # determine bin stats
        bin_stats1, total_uniq_binned_seqs1, tota_uniq_binned_bases1, num_repeats1 = self._binning_stats(
            bins1, seq_lens)
        bin_stats2, total_uniq_binned_seqs2, tota_uniq_binned_bases2, num_repeats2 = self._binning_stats(
            bins2, seq_lens)

        # sort bins by size
        bin_stats1 = sorted(bin_stats1.items(),
                            key=lambda x: x[1][1], reverse=True)
        bin_stats2 = sorted(bin_stats2.items(),
                            key=lambda x: x[1][1], reverse=True)

        # report summary results
        print()
        print('Assembled sequences = %d (%.2f Mbp)' %
              (len(seqs), total_bases / 1e6))
        print('  No. seqs > 1 kbp = %d (%.2f Mbp)' %
              (num_seq1K, total_bases1K / 1e6))
        print('  No. seqs > 5 kbp = %d (%.2f Mbp)' %
              (num_seq5K, total_bases5K / 1e6))
        print()
        print('Binning statistics:')
        print('  1) No. bins: %s, No. binned seqs: %d (%.2f%%), No. binned bases: %.2f Mbp (%.2f%%), No. seqs in multiple bins: %d'
              % (len(bins1),
                 total_uniq_binned_seqs1,
                 total_uniq_binned_seqs1 * 100 / len(seqs),
                 tota_uniq_binned_bases1 / 1e6,
                 tota_uniq_binned_bases1 * 100 / total_bases,
                 num_repeats1))
        print('  2) No. bins: %s, No. binned seqs: %d (%.2f%%), No. binned bases: %.2f Mbp (%.2f%%), No. seqs in multiple bins: %d'
              % (len(bins2),
                 total_uniq_binned_seqs2,
                 total_uniq_binned_seqs2 * 100 / len(seqs),
                 tota_uniq_binned_bases2 / 1e6,
                 tota_uniq_binned_bases2 * 100 / total_bases,
                 num_repeats2))
        print()

        # output report
        fout = open(output_file, 'w')
        for data in bin_stats2:
            fout.write('\t' + data[0])
        fout.write(
            '\tUnbinned\tNo. Sequences\tNo. Bases (Mbp)\tBest Match\tBases in Common (%)\tSequences in Common (%)\n')

        max_bases_in_common2 = defaultdict(int)
        max_seqs_in_common2 = defaultdict(int)
        best_matching_bins2 = {}
        binned_seqs2 = defaultdict(set)
        for data1 in bin_stats1:
            bin_id1 = data1[0]
            fout.write(bin_id1)

            seqs1 = bins1[bin_id1]

            max_bases_in_common = 0
            max_seqs_in_common = 0
            best_matching_bin = 'n/a'
            binned_seqs = set()
            for data2 in bin_stats2:
                bin_id2 = data2[0]
                seqs2 = bins2[bin_id2]

                seqs_in_common = seqs1.intersection(seqs2)
                binned_seqs.update(seqs_in_common)
                num_seqs_in_common = len(seqs_in_common)
                fout.write('\t' + str(num_seqs_in_common))

                bases_in_common = 0
                for seq_id in seqs_in_common:
                    bases_in_common += seq_lens[seq_id]

                if bases_in_common > max_bases_in_common:
                    max_bases_in_common = bases_in_common
                    max_seqs_in_common = num_seqs_in_common
                    best_matching_bin = bin_id2

                if bases_in_common > max_bases_in_common2[bin_id2]:
                    max_bases_in_common2[bin_id2] = bases_in_common
                    max_seqs_in_common2[bin_id2] = num_seqs_in_common
                    best_matching_bins2[bin_id2] = bin_id1

                binned_seqs2[bin_id2].update(seqs_in_common)
            fout.write('\t{}\t{}\t{:.2f}\t{}\t{:.2f}\t{:.2f}\n'.format(
                len(seqs1) - len(binned_seqs),
                data1[1][0],
                data1[1][1] /
                1e6,
                best_matching_bin,
                max_bases_in_common *
                100 /
                data1[1][1],
                max_seqs_in_common *
                100 /
                data1[1][0],
            ))

        fout.write('Unbinned')
        for data in bin_stats2:
            binId = data[0]
            fout.write('\t%d' % (len(bins2[binId]) - len(binned_seqs2[binId])))
        fout.write('\n')

        fout.write('No. Sequences')
        for data in bin_stats2:
            fout.write('\t%d' % data[1][0])
        fout.write('\n')

        fout.write('No. Bases (Mbp)')
        for data in bin_stats2:
            fout.write('\t%.2f' % (data[1][1] / 1e6))
        fout.write('\n')

        fout.write('Best Match')
        for data in bin_stats2:
            binId = data[0]
            fout.write('\t%s' % best_matching_bins2.get(binId, 'n/a'))
        fout.write('\n')

        fout.write('Bases in Common (%)')
        for data in bin_stats2:
            binId = data[0]
            fout.write('\t%.2f' %
                       (max_bases_in_common2[binId] * 100 / data[1][1]))
        fout.write('\n')

        fout.write('Sequences in Common (%)')
        for data in bin_stats2:
            binId = data[0]
            fout.write('\t%.2f' %
                       (max_seqs_in_common2[binId] * 100 / data[1][0]))
        fout.write('\n')

        fout.close()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import shutil
import logging
from collections import defaultdict

from unitem.utils import check_dependencies
from unitem.external.prodigal import Prodigal
from unitem.external.tigrfam_search import TigrfamSearch
from unitem.external.pfam_search import PfamSearch
from unitem.external.tophits import TopHitFile
from unitem.markers import Markers
from unitem.defaults import (BINNING_METHOD_DIR,
                             TIGRFAM_TOP_HIT_SUFFIX,
                             PFAM_TOP_HIT_SUFFIX)


class Profile():
    """Profile genomes across different binning methods."""

    def __init__(self, cpus):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

        check_dependencies(['prodigal', 'hmmsearch'])

        self.logger.info(f'Using Prodigal {Prodigal.version()}.')
        self.logger.info(f'Using hmmsearch {TigrfamSearch.version()}.')

        self.cpus = cpus

    def _genome_quality(self, gids, result_dir, tigr_search, pfam_search):
        """Determine genome quality based on presence/absence of marker genes."""

        pfam_bac_ms, pfam_ar_ms = pfam_search.get_marker_genes()
        tigr_bac_ms, tigr_ar_ms = tigr_search.get_marker_genes()

        bac_ms = pfam_bac_ms.union(tigr_bac_ms)
        ar_ms = pfam_ar_ms.union(tigr_ar_ms)

        fout_bac = open(os.path.join(result_dir, 'bac_hits.tsv'), 'w')
        fout_bac.write('Genome ID')
        for ms in bac_ms:
            fout_bac.write(f'\t{ms}')
        fout_bac.write('\n')

        fout_ar = open(os.path.join(result_dir, 'ar_hits.tsv'), 'w')
        fout_ar.write('Genome ID')
        for ms in ar_ms:
            fout_ar.write(f'\t{ms}')
        fout_ar.write('\n')

        genome_qual = {}
        for gid in gids:
            hmm_hits = defaultdict(list)

            # get top hits to PFAM HMMs
            pfam_top_hit_file = os.path.join(
                result_dir, f'{gid}{PFAM_TOP_HIT_SUFFIX}')
            pfam_top_hits = TopHitFile(pfam_top_hit_file)
            pfam_top_hits.read()

            for gene_id, cur_hit in pfam_top_hits.iter_hits():
                scaffold_id = gene_id.rpartition('_')[0]
                hmm_hits[cur_hit.hmm_id].append(scaffold_id)

            # get top hits to TIGRFAM HMMs
            tigr_top_hit_file = os.path.join(
                result_dir, f'{gid}{TIGRFAM_TOP_HIT_SUFFIX}')
            tigr_top_hits = TopHitFile(tigr_top_hit_file)
            tigr_top_hits.read()

            for gene_id, cur_hit in tigr_top_hits.iter_hits():
                scaffold_id = gene_id.rpartition('_')[0]
                hmm_hits[cur_hit.hmm_id].append(scaffold_id)

            bac_comp, bac_cont = Markers.estimate_comp_cont(
                hmm_hits,
                bac_ms
            )

            ar_comp, ar_cont = Markers.estimate_comp_cont(
                hmm_hits,
                ar_ms
            )

            if bac_comp + bac_cont > ar_comp + ar_cont:
                genome_qual[gid] = ('Bacteria', bac_comp, bac_cont)

                fout_bac.write(f'{gid}')
                for ms in bac_ms:
                    fout_bac.write(f'\t{len(hmm_hits[ms])}')
                fout_bac.write('\n')
            else:
                genome_qual[gid] = ('Archaea', bac_comp, bac_cont)

                fout_ar.write(f'{gid}')
                for ms in ar_ms:
                    fout_ar.write(f'\t{len(hmm_hits[ms])}')
                fout_ar.write('\n')

        fout_bac.close()
        fout_ar.close()

        return genome_qual

    def _report_genome_quality(self, genome_quality, output_dir):
        """Summarize quality of genomes."""

        # create table for each binning method
        for bm in genome_quality:
            table = os.path.join(output_dir, bm + '_quality.tsv')
            fout = open(table, 'w')
            fout.write(
                'Genome ID\tMarker Set Domain\tCompleteness (%)\tContamination (%)\tQuality\n')
            for gid in genome_quality[bm]:
                domain, comp, cont = genome_quality[bm][gid]
                fout.write('{}\t{}\t{:.2f}\t{:.2f}\t{:.2f}\n'.format(
                           gid,
                           domain,
                           comp,
                           cont,
                           comp-5*cont))
            fout.close()

        # report global results file
        summary = {}
        for bm in genome_quality:
            comp_cont_count = defaultdict(lambda: defaultdict(int))
            quality_count = defaultdict(int)
            total_comp = 0
            total_cont = 0
            total_q = 0
            for _bid, (domain, comp, cont) in genome_quality[bm].items():
                quality = comp - 5*cont

                for test_comp in [90, 80, 70]:
                    for test_cont in [5, 10]:
                        if comp >= test_comp and cont <= test_cont:
                            comp_cont_count[test_comp][test_cont] += 1

                for test_q in [90, 70, 50]:
                    if quality >= test_q:
                        quality_count[test_q] += 1

                total_comp += comp
                total_cont += cont
                total_q += quality

                summary[bm] = (comp_cont_count, quality_count,
                               total_comp, total_cont, total_q)

        fout = open(os.path.join(output_dir, 'bin_quality_summary.tsv'), 'w')
        fout.write('\t' + '\t'.join(sorted(summary)) + '\n')

        for comp in [90, 80, 70]:
            for cont in [5, 10]:
                fout.write('Comp. >= %d, cont. <= %d' % (comp, cont))
                for bm in sorted(summary):
                    fout.write('\t%d' % summary[bm][0][comp][cont])
                fout.write('\n')

        for q in [90, 70, 50]:
            fout.write('Quality >= %d' % q)
            for bm in sorted(summary):
                fout.write('\t%d' % summary[bm][1][q])
            fout.write('\n')

        fout.write('Total compl.')
        for bm in sorted(summary):
            fout.write('\t%.1f' % summary[bm][2])
        fout.write('\n')

        fout.write('Total cont.')
        for bm in sorted(summary):
            fout.write('\t%.1f' % summary[bm][3])
        fout.write('\n')

        fout.write('Total quality')
        for bm in sorted(summary):
            fout.write('\t%.1f' % summary[bm][4])
        fout.write('\n')

        fout.close()

    def run(self, bin_dirs, marker_dir, output_dir):
        """Profile genomes in each bin directory.

        Parameters
        ----------
        bin_dirs : list of str
            Directories containing bins from different binning methods.
        marker_dir : str
            Directory containing Pfam and TIGRfam marker gene data.
        output_dir : str
            Output directory.
        """

        self.logger.info(f'Profiling genomes in {len(bin_dirs)} directories.')

        num_processed = 0
        genome_quality = defaultdict(lambda: dict)
        prodigal = Prodigal(self.cpus)
        for method_id, (bin_dir, bin_ext) in bin_dirs.items():
            num_processed += 1
            self.logger.info('Profiling {} ({} of {}):'.format(
                method_id,
                num_processed,
                len(bin_dirs)))

            # path to genomic FASTA files
            genomes = {}
            for gf in os.listdir(bin_dir):
                if not gf.endswith(bin_ext):
                    continue

                gid = gf.replace(f'.{bin_ext}', '')
                genomes[gid] = os.path.join(bin_dir, gf)

            # identify genes using Prodigal
            self.logger.info(
                f' - identifying genes in {len(genomes):,} genomes using Prodigal')

            method_dir = os.path.join(output_dir,
                                      BINNING_METHOD_DIR,
                                      method_id)
            tmp_marker_gene_dir = os.path.join(method_dir, 'tmp_marker_genes')

            prodigal_files = prodigal.run(genomes, tmp_marker_gene_dir)

            # annotated genes against TIGRFAM database
            self.logger.info(' - identifying TIGRFAM markers using HMMER')
            aa_gene_files = [prodigal_files[gid]['aa_gene_path']
                             for gid in prodigal_files.keys()]
            tigr_search = TigrfamSearch(marker_dir, self.cpus)
            tigr_search.run(aa_gene_files, tmp_marker_gene_dir, method_dir)

            # annotate genes against Pfam database
            self.logger.info(' - identifying Pfam markers using HMMER')
            pfam_search = PfamSearch(marker_dir, self.cpus)
            pfam_search.run(aa_gene_files, tmp_marker_gene_dir,  method_dir)

            genome_quality[method_id] = self._genome_quality(
                genomes,
                method_dir,
                tigr_search,
                pfam_search)

            shutil.rmtree(tmp_marker_gene_dir)

        self._report_genome_quality(genome_quality, output_dir)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
from collections import Counter

from unitem.utils import (make_sure_path_exists,
                          check_dir_exists,
                          check_file_exists)
from unitem.bin_tools import BinTools
from unitem.bin import Bin
from unitem.profile import Profile
from unitem.ensemble import Ensemble


class OptionsParser():
    def __init__(self):
        """Initialization"""

        self.logger = logging.getLogger('timestamp')

    def _bin_extension(self, bin_dir):
        """Determine extension of bins."""

        exts = []
        for f in os.listdir(bin_dir):
            f_split = f.split('.')

            if len(f_split) > 1:
                ext = f_split[-1]
                if ext in ['fa', 'fna', 'fasta']:
                    exts.append(ext)

        if len(exts) == 0:
            return None, None

        ext, count = Counter(exts).most_common(1)[0]
        return ext, count

    def _bin_dirs(self, options):
        """Get directories with bins from different binners."""

        bin_dirs = {}
        if hasattr(options, 'bin_dirs') and options.bin_dirs:
            for d in options.bin_dirs:
                check_dir_exists(d)
                method_id = os.path.basename(os.path.normpath(d))
                bin_ext, count = self._bin_extension(d)
                if not bin_ext:
                    self.logger.warning(
                        f'No bins identified for {method_id} in {d}.')
                else:
                    bin_dirs[method_id] = (d, bin_ext)
                    self.logger.info(
                        f"Processing {count} genomes from {method_id} with extension '{bin_ext}'.")

        if hasattr(options, 'bin_file') and options.bin_file:
            check_file_exists(options.bin_file)
            for line in open(options.bin_file):
                if line.strip():
                    tokens = [token.strip() for token in line.split('\t')]
                    if len(tokens) != 2:
                        self.logger.warning(
                            f"Skipping invalid line: {line.strip()}")
                        continue

                    method_id = tokens[0]
                    d = tokens[1]
                    check_dir_exists(d)
                    bin_ext, count = self._bin_extension(d)
                    if not bin_ext:
                        self.logger.warning(
                            f'No bins identified for {method_id} in {d}.')
                    else:
                        bin_dirs[method_id] = (d, bin_ext)
                        self.logger.info(
                            f"Processing {count} genomes from {method_id} with extension '{bin_ext}'.")

        return bin_dirs

    def bin(self, options):
        """Bin command"""

        check_file_exists(options.assembly_file)
        make_sure_path_exists(options.output_dir)

        bin = Bin(options.assembly_file,
                  options.output_dir,
                  options.min_contig_len,
                  options.cpus)
        bin.check_on_path(options)
        bin.coverage(options.bam_files, options.cov_file)
        bin.run(options)

        self.logger.info(
            f"UniteM 'bin' results written to: {options.output_dir}")

    def profile(self, options):
        """Profile command"""

        make_sure_path_exists(options.output_dir)

        bin_dirs = self._bin_dirs(options)

        profile = Profile(options.cpus)
        profile.run(bin_dirs,
                    options.marker_dir,
                    options.output_dir)

        self.logger.info(
            f"UniteM 'profile' results written to: {options.output_dir}")

    def consensus(self, options):
        """Consensus command"""

        check_dir_exists(options.profile_dir)
        make_sure_path_exists(options.output_dir)

        bin_dirs = self._bin_dirs(options)

        e = Ensemble(options.bin_prefix)
        e.run(options.profile_dir,
              bin_dirs,
              options.marker_dir,
              options.weight,
              options.sel_min_quality,
              options.sel_min_comp,
              options.sel_max_cont,
              options.remove_perc,
              options.add_perc,
              options.add_matches,
              False,  # perform consensus bin selection
              False,  # perform unanimous bin selection
              options.report_min_quality,
              options.simple_headers,
              options.output_dir)

        self.logger.info(
            f"UniteM 'consensus' results written to: {options.output_dir}")

    def greedy(self, options):
        """Greedy command"""

        check_dir_exists(options.profile_dir)
        make_sure_path_exists(options.output_dir)

        bin_dirs = self._bin_dirs(options)

        e = Ensemble(options.bin_prefix)
        e.run(options.profile_dir,
              bin_dirs,
              options.marker_dir,
              options.weight,
              options.sel_min_quality,
              options.sel_min_comp,
              options.sel_max_cont,
              None,
              None,
              None,
              True,   # perform greedy bin selection
              False,  # perform unanimous bin selection
              options.report_min_quality,
              options.simple_headers,
              options.output_dir)

        self.logger.info(
            f"UniteM 'greedy' results written to: {options.output_dir}")

    def unanimous(self, options):
        """Unanimous command"""

        check_dir_exists(options.profile_dir)
        make_sure_path_exists(options.output_dir)

        bin_dirs = self._bin_dirs(options)

        e = Ensemble(options.bin_prefix)
        e.run(options.profile_dir,
              bin_dirs,
              options.marker_dir,
              options.weight,
              options.sel_min_quality,
              options.sel_min_comp,
              options.sel_max_cont,
              None,
              None,
              None,
              False,  # perform greedy bin selection
              True,   # perform unanimous bin selection
              options.report_min_quality,
              options.simple_headers,
              options.output_dir)

        self.logger.info(
            f"UniteM 'unanimous' results written to: {options.output_dir}")

    def compare(self, options):
        """Compare command"""

        check_dir_exists(options.bin_dir1)
        check_dir_exists(options.bin_dir2)

        bt = BinTools()
        bin_files1 = bt.bin_files(options.bin_dir1, options.extension1)
        bin_files2 = bt.bin_files(options.bin_dir2, options.extension2)

        bt.compare(bin_files1, bin_files2,
                   options.assembly_file, options.output_file)

        self.logger.info("UniteM 'compare' results written to: %s" %
                         options.output_file)

    def unique(self, options):
        """Unique command"""

        check_dir_exists(options.bin_dir)

        bt = BinTools()
        bin_files = bt.bin_files(options.bin_dir, options.extension)
        bt.unique(bin_files)

    def parse_options(self, options):
        """Parse user options and call the correct pipeline(s)"""

        if(options.subparser_name == 'bin'):
            self.bin(options)
        elif(options.subparser_name == 'profile'):
            self.profile(options)
        elif(options.subparser_name == 'consensus'):
            self.consensus(options)
        elif(options.subparser_name == 'greedy'):
            self.greedy(options)
        elif(options.subparser_name == 'unanimous'):
            self.unanimous(options)
        elif(options.subparser_name == 'compare'):
            self.compare(options)
        elif(options.subparser_name == 'unique'):
            self.unique(options)
        else:
            self.logger.error(
                f'Unknown UniteM command: {options.subparser_name}\n')
            sys.exit(1)

        return 0



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import subprocess
import multiprocessing as mp
import inspect

from unitem.external.tophits import TopHitTigrFile
from unitem.external.prodigal import AA_GENE_FILE_SUFFIX

TIGRFAM_SUFFIX = "_tigrfam.tsv"
TIGRFAM_OUT = '_tigrfam.out'


class TigrfamException(Exception):
    """Exception for errors associated with running HMMER with TIGRfam marker genes."""

    def __init__(self, message=''):
        Exception.__init__(self, message)


class TigrfamSearch(object):
    """Runs TIGRFAM HMMs over a set of genomes."""

    def __init__(self, marker_dir, cpus):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

        self.cpus = cpus

        # file with TIGRfam HMMs
        if marker_dir is None:
            hmm_file = os.path.join(TigrfamSearch.hmm_dir(), 'tigrfam.hmm')
            self.marker_dir = TigrfamSearch.hmm_dir()
            self.tigrfam_hmms = os.path.abspath(hmm_file)
        else:
            self.marker_dir = marker_dir
            self.tigrfam_hmms = os.path.abspath(
                os.path.join(marker_dir, 'tigrfam.hmm'))

        self.logger.info(f' - using TIGRfam HMMs in: {self.marker_dir}')

    @staticmethod
    def version():
        """ get HMMER version."""
        try:
            proc = subprocess.Popen(['hmmsearch', '-h'],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    encoding='utf-8')

            output, _ = proc.communicate()
            for line in output.split('\n'):
                if line.startswith('# HMMER'):
                    version = line.split(';')[0].replace('# HMMER', '').strip()
                    return version
            return "(version unavailable)"
        except:
            return "(version unavailable)"

    @staticmethod
    def hmm_dir():
        """Get directory with TIGRAM HMM model files."""

        src_file_path = os.path.split(inspect.getfile(lambda: None))[0]
        hmm_dir = os.path.abspath(
            os.path.join(src_file_path, '..', '..', 'markers'))

        return hmm_dir

    def get_marker_genes(self):
        """Get bacterial and archaeal PFAM marker genes."""

        bac_ms = set()
        with open(os.path.join(self.marker_dir, 'tigr_bac.lst')) as f:
            for line in f:
                bac_ms.add(line.strip())

        ar_ms = set()
        with open(os.path.join(self.marker_dir, 'tigr_ar.lst')) as f:
            for line in f:
                ar_ms.add(line.strip())

        return bac_ms, ar_ms

    def _top_hit(self, tigrfam_file, output_dir):
        """Determine top hits to TIGRFAMs.

        A gene is assigned to a single TIGRFAM
        family. This will be the top hit among
        all TIGRFAM HMMs and pass the threshold
        for the HMM.

        Parameters
        ----------
        tigrfam_file : str
            Name of file containing hits to TIGRFAM HMMs.
        """

        filename = os.path.split(tigrfam_file)[1]
        gid = filename.replace(TIGRFAM_SUFFIX, '')
        tophit_file = TopHitTigrFile(output_dir, gid)

        # populate the top hit file
        with open(tigrfam_file, 'r') as fh_tigrfam:
            for line in fh_tigrfam:
                if line[0] == '#':
                    continue

                line_split = line.split()
                gene_id = line_split[0]
                hmm_id = line_split[3]
                evalue = float(line_split[4])
                bitscore = float(line_split[5])
                tophit_file.add_hit(gene_id, hmm_id, evalue, bitscore)

        # write the top-hit file to disk
        tophit_file.write()

    def _worker(self, gene_dir, output_dir, queue_in, queue_out):
        """Process each genome in parallel."""

        while True:
            gene_file = queue_in.get(block=True, timeout=None)
            if gene_file is None:
                break

            filename = os.path.split(gene_file)[1]
            output_hit_file = os.path.join(gene_dir, filename.replace(AA_GENE_FILE_SUFFIX,
                                                                      TIGRFAM_SUFFIX))

            hmmsearch_out = os.path.join(gene_dir, filename.replace(AA_GENE_FILE_SUFFIX,
                                                                    TIGRFAM_OUT))

            if os.path.exists(output_hit_file) and os.path.exists(hmmsearch_out):
                # use previously calculated results
                queue_out.put(gene_file)
                continue

            # identify TIGRfam marker genes
            cmd = ['hmmsearch', '--noali', '--notextw', '--cut_nc',
                   '-o', hmmsearch_out,
                   '--tblout', output_hit_file,
                   '--cpu', str(self.cpus_per_genome),
                   self.tigrfam_hmms, gene_file]
            p = subprocess.Popen(cmd, stderr=subprocess.PIPE, encoding='utf-8')
            _stdout, stderr = p.communicate()

            if p.returncode != 0:
                self.logger.error(
                    f'Non-zero exit code returned when running hmmsearch: {stderr}')
                self.logger.error(f'Skipping {filename}.')

                return None

            # identify top hit for each gene
            self._top_hit(output_hit_file, output_dir)

            # allow results to be processed or written to file
            queue_out.put(gene_file)

    def _writer(self, num_genomes, writer_queue):
        """Track progress."""

        num_processed = 0
        while True:
            data = writer_queue.get(block=True, timeout=None)
            if data is None:
                break

            num_processed += 1
            status = f' -> processed {num_processed:,} of {num_genomes:,} genomes'
            sys.stdout.write(f'{status}\r')
            sys.stdout.flush()

    def run(self, gene_files, gene_dir, output_dir):
        """Annotate genes with TIGRFAM HMMs.

        Parameters
        ----------
        gene_files : iterable
            Gene files in FASTA format to process.
        """

        if len(gene_files) == 0:
            self.logger.error('No genomes to process.')
            sys.exit(1)

        self.cpus_per_genome = max(1, int(self.cpus / len(gene_files)))

        # populate worker queue with data to process
        worker_queue = mp.Queue()
        writer_queue = mp.Queue()

        for f in gene_files:
            worker_queue.put(f)

        for _ in range(self.cpus):
            worker_queue.put(None)

        try:
            worker_proc = [mp.Process(target=self._worker, args=(gene_dir,
                                                                 output_dir,
                                                                 worker_queue,
                                                                 writer_queue)) for _ in range(self.cpus)]

            write_proc = mp.Process(target=self._writer, args=(len(gene_files),
                                                               writer_queue))

            write_proc.start()

            for p in worker_proc:
                p.start()

            for p in worker_proc:
                p.join()

                if p.exitcode != 0:
                    raise TigrfamException(
                        'HMMER returned non-zero exit code.')

            writer_queue.put(None)
            write_proc.join()
        except Exception:
            for p in worker_proc:
                p.terminate()

            write_proc.terminate()
            raise
